#!/usr/bin/env python3
"""
Setup script for KiCadFiles.

This file is mainly for compatibility. The actual configuration
is in pyproject.toml following PEP 518.
"""

from setuptools import setup

if __name__ == "__main__":
    setup()
